#!/bin/sh -u

readonly THIS_PATH="$(readlink -e "$(dirname "$0")")"

. "${THIS_PATH}/include/utils.inc"
. "${THIS_PATH}/include/tools.inc"
. "${THIS_PATH}/include/is-product-ready.inc"
. "${THIS_PATH}/include/upgrade-state.inc"
. "${THIS_PATH}/include/offline-event.inc"
. "${THIS_PATH}/include/ini-file.inc"

readonly WDSERVER_BIN='/opt/kaspersky/kesl/libexec/wdserver'
readonly PRODUCT_NAME='kesl'
readonly PRODUCT_VERSION='10.1.2.9000'
readonly PRODUCT_BIN='/opt/kaspersky/kesl/libexec/kesl'
readonly WDSERVER_LOG='/var/log/kaspersky/kesl/kesl_wdserver.log'
readonly WDSERVER_PID='/var/run/wdserver.pid'
readonly LAUNCHER_PID='/var/run/kesl_launcher.pid'
readonly LAUNCHER_LOG='/var/log/kaspersky/kesl/kesl_launcher.log'
readonly LAUNCHER_FILEPATH='/opt/kaspersky/kesl/libexec/kesl_launcher.sh'
readonly PRODUCT_CONTROL_BIN='/opt/kaspersky/kesl/bin/kesl-control'
readonly VAR_DIRECTORY='/var/opt/kaspersky/kesl/'
readonly SHARED_INIT_STORAGE_DIRECTORY='/opt/kaspersky/kesl/shared/init/storage/'
readonly VAR_STORAGE_DIRECTORY='/var/opt/kaspersky/kesl/private/storage/'
readonly BASES_INIT_DATA_DATA_DIRECTORY='/opt/kaspersky/kesl/shared/init/updates/data/data/'
readonly BASES_DATA_DIRECTORY='/var/opt/kaspersky/kesl/common/updates/data/data/'
readonly SC_KLNAGENT_BIN='/opt/kaspersky/klnagent64/sbin/klnagent'
readonly READY_MARKER_FILEPATH='/var/run/kesl.mrk'
readonly INI_FILEPATH='/var/opt/kaspersky/kesl/common/kesl.ini'
readonly MANIFEST_FILEPATH='/opt/kaspersky/kesl/doc/manifest.ini'
readonly MONITOR_OAS_VERSION_FILEPATH='/proc/kav4fs_oas.version'
readonly EVENTS_STORAGE_FILEPATH='/var/opt/kaspersky/kesl/private/storage/events.db'
readonly INSTALL_FILEPATH='/opt/kaspersky/kesl/libexec/install'
readonly INSTALL_LOG_FILEPATH='/var/opt/kaspersky/kesl/private/install.log'
readonly UNINSTALL_FILEPATH='/opt/kaspersky/kesl/libexec/uninstall'
readonly CURRENT='/var/opt/kaspersky/kesl/install-current'
readonly UPGRADE='/var/opt/kaspersky/kesl/install-upgrade'
readonly PREVIOUS='/var/opt/kaspersky/kesl/install-previous'
readonly FAILED='/var/opt/kaspersky/kesl/install-failed'
readonly CURRENT_NAME='install-current'
readonly UPGRADE_NAME='install-upgrade'
readonly PREVIOUS_NAME='install-previous'
readonly FAILED_NAME='install-failed'

readonly CRYPTO_SSL_WORKING_COPY_FILEPATH='/opt/kaspersky/kesl/lib64/updatable_modules/libcrypto_ssl_working_copy.so'
readonly CRYPTO_SSL_SYMLINK_NEW_FILEPATH='/opt/kaspersky/kesl/lib64/updatable_modules/libcrypto_ssl_new'
readonly CRYPTO_SSL_PREV_FILEPATH='/opt/kaspersky/kesl/lib64/updatable_modules/previous/libcrypto_ssl_working_copy.so'
readonly CRYPTO_SSL_SYMLINK_FILEPATH='/opt/kaspersky/kesl/lib64/libcrypto_ssl_1_1.so'
readonly CRYPTO_SSL_FILEPATH='/opt/kaspersky/kesl/lib64/libcrypto_ssl_working_copy.so'

readonly PREPARE_MODULE_SCRIPT_FILEPATH='/opt/kaspersky/kesl/libexec/prepare_module.sh'
readonly BACKUP_DIRECTORY="/var/opt/kaspersky/kesl/private/storage-backup"
readonly BACKUP_CURRENT_LNK="/var/opt/kaspersky/kesl/private/storage-backup.current"
readonly DEFAULT_FILE_MODE_CREATION_MASK='0077'
readonly LOGDATEFMT='+%F %T.%3N%:z'

readonly LINUX_KERNEL="$(uname -r)"

readonly PRODUCT_START_DELAY=180

readonly SERVER_REBOOT_DELAY=0

readonly WDRESTART=5

readonly WDSERVER_LOG_LEVEL=0

readonly WDSERVER_SOCK='/var/run/wdserver.sock'

readonly WDTERM_SIGNAL=2

readonly WDSERVER_MAX_MEM='90%'
readonly WDSERVER_MAX_SWAP='100%'

readonly WDSERVER_FLAGS="\
--trace-level ${WDSERVER_LOG_LEVEL} \
--trace-file ${WDSERVER_LOG} \
--n-restart ${WDRESTART} \
--master-timeout ${PRODUCT_START_DELAY} \
--reboot-timeout ${SERVER_REBOOT_DELAY} \
--wdsock ${WDSERVER_SOCK} \
--max-used-mem ${WDSERVER_MAX_MEM} \
--max-used-swap ${WDSERVER_MAX_SWAP} \
--term-signal ${WDTERM_SIGNAL} \
-p ${WDSERVER_PID} \
-d"

BACKUP_CURRENT=''
UPGRADE_STATE=''
PATCH_VERSION=''
OLD_VERSION=''
BACKUP_NEW_UNIQUE_DIR=''

Log()
{
    Echo "[$(date "${LOGDATEFMT}")] $*"
}

Trace()
{
    Log "|trace| $*"
}

GetBackupDir()
{
    local outVar="${1:?}"

    if [ ! -L "${BACKUP_CURRENT_LNK}" ]; then
        return 0
    fi

    local current
    current="$(readlink -n "${BACKUP_CURRENT_LNK}")"

    local rc
    if [ "${current}" = "${SHARED_INIT_STORAGE_DIRECTORY}" ]; then
        rc=1
    elif [ -z "${current}" ] || [ ! -d "${current}" ]; then
        rc=1
        current="${SHARED_INIT_STORAGE_DIRECTORY}"
    else
        rc=2
    fi

    eval "${outVar}='${current}'"

    return ${rc}
}

MakeUniqueDir()
{
    Trace "MakeUniqueDir()"


    if [ ! -d "${BACKUP_DIRECTORY}" ]; then
        mkdir -v "${BACKUP_DIRECTORY}"
    fi

    local tmp
    tmp="$(mktemp -d --tmpdir="${BACKUP_DIRECTORY}" "$(date '+%FT%T%:z') (XXXXXX)")"
    local res=$?

    if [ $res -ne 0 ] || [ -z "${tmp}" ]; then
        Log "Cannot create temporary directory in dir '${BACKUP_DIRECTORY}'"
    else
        BACKUP_NEW_UNIQUE_DIR="${tmp}"
    fi

    return $res
}

RemoveNewUniqDir()
{
    Trace "RemoveNewUniqDir()"

    local path="${BACKUP_NEW_UNIQUE_DIR}"

    Log "Removing temporary directory '${path}'"

    if [ ! -z "${path}" ]; then
        rm -rfv "${path}"
    fi

    BACKUP_NEW_UNIQUE_DIR=""
}

ClearOldBackupDirs()
{
    Trace "ClearOldBackupDirs()"


    local currentName
    currentName="$(basename "${BACKUP_NEW_UNIQUE_DIR}")" || currentName=""
    find "${BACKUP_DIRECTORY}" -mindepth 1 -maxdepth 1 -not -name "${currentName}" \
        -exec rm -rfv {} \;
}

PrepareBackup()
{
    Trace "PrepareBackup()"

    if MakeUniqueDir; then
        Log "Preparing backup in '${BACKUP_NEW_UNIQUE_DIR}' directory ..."

        cp -frpv -t "${BACKUP_NEW_UNIQUE_DIR}" "${VAR_STORAGE_DIRECTORY}/."
        local res=$?
        if [ ${res} -ne 0 ]; then
            Log "Preparing backup completed with error: ${res}"
            RemoveNewUniqDir
            return ${res}
        fi

        Log "Preparing backup successfully completed"
    fi
}

CommitBackup()
{
    Trace "CommitBackup()"

    local res

    if [ -z "${BACKUP_NEW_UNIQUE_DIR}" ]; then
        Log "Commit backup to init directory '${SHARED_INIT_STORAGE_DIRECTORY}' ..."
        ln -Tsfv "${SHARED_INIT_STORAGE_DIRECTORY}" "${BACKUP_CURRENT_LNK}"
        res=$?
    else
        Log "Commit backup to directory '${BACKUP_NEW_UNIQUE_DIR}' ..."
        ln -Tsfv "${BACKUP_NEW_UNIQUE_DIR}" "${BACKUP_CURRENT_LNK}"
        res=$?
        if [ ${res} -ne 0 ]; then
            RemoveNewUniqDir
        else
            ClearOldBackupDirs
            BACKUP_NEW_UNIQUE_DIR=""
        fi
    fi

    if [ ${res} -ne 0 ]; then
        Log "Commit backup error: cannot create 'current' symlink: ${res}"
    else
        Log "Commit backup OK"
    fi

    return ${res}
}

RestoreCryptoSSL()
{
    Trace "RestoreCryptoSSL()"

    local restore_dir="$1"
    local init_dir="${SHARED_INIT_STORAGE_DIRECTORY}"
    local prev="${CRYPTO_SSL_PREV_FILEPATH}"
    local current="${CRYPTO_SSL_SYMLINK_FILEPATH}"

    if [ "${restore_dir}" = "${init_dir}" ] || [ ! -e "${prev}" ] || \
       [ "$(readlink -e "${prev}")" = "$(readlink -e "${current}")" ]; then
        local initial="${CRYPTO_SSL_FILEPATH}"
        ln -sfT "${initial}" "${current}"
    else
        ln -sfT "${prev}" "${current}"
    fi

    return 0
}

RestoreFromBackup()
{
    Trace "RestoreFromBackup()"

    local current="$1"

    if [ "${current}" = "" ]; then
        current="${SHARED_INIT_STORAGE_DIRECTORY}"
    fi

    Log "Restoring from backup located at '${current}'"

    rm -rfv "${VAR_STORAGE_DIRECTORY}/"*
    cp -frpv -t "${VAR_STORAGE_DIRECTORY}" "${current}/."

    RestoreCryptoSSL "${current}"

    Log "Restore completed"

    return 0
}

PopulateDataDirectory()
{
    mkdir -p "${BASES_DATA_DIRECTORY}"
    cp -n "${BASES_INIT_DATA_DATA_DIRECTORY}"* "${BASES_DATA_DIRECTORY}"
}

Stop()
{
    Trace "Stop()"

    local wdPid
    wdPid="$(head "${WDSERVER_PID}")"
    if [ ! -z "${wdPid}" ]; then
        while IsProductStarting; do
            Trace "Product is still running: try to kill the watchdog by SIGINT: wdPid="${wdPid}""
            kill -INT "${wdPid}" || break
            sleep 1
        done

        Trace "kill -9 for wdPid=${wdPid}"
        kill -9 "${wdPid}" > /dev/null 2>&1
        Trace "wait for wdPid=${wdPid}"
        wait "${wdPid}" > /dev/null 2>&1
        Trace "wait done"
    fi

    Trace "Stop() done"

    exit 0
}

IsFanotifyOn()
{
    local useFanotify=''
    if ! ReadIni "${INI_FILEPATH}" 'General' 'UseFanotify' useFanotify; then
        return 1
    fi

    if ! echo "${useFanotify}" | grep -iE '^(true|yes|1)$' >/dev/null; then
        return 1
    fi

    return 0
}

IsKernelModuleLoaded()
{
    test -f "${MONITOR_OAS_VERSION_FILEPATH}"
    return $?
}

PrepareKernelModule()
{
    Trace "PrepareKernelModule(): $@"

    local atPath="${1:?}"
    shift

    if ! "${atPath}/${PREPARE_MODULE_SCRIPT_FILEPATH}" "$@"; then
        return 1
    fi

    return 0
}

RemoveKernelModuleFiles()
{
    rm -f "/lib/modules/${LINUX_KERNEL}/misc/kav4fs_oas.ko"
    rm -f "/lib/modules/${LINUX_KERNEL}/misc/redirfs.ko"
}

RevokePatchApproval()
{
    local folder="$1"

    local mnfstIniFile="${folder}/${MANIFEST_FILEPATH}"

    local mnfstPatchId=''
    ReadIni "${mnfstIniFile}" '' 'PatchId' mnfstPatchId || return $?

    local section='AcceptedAgreement'
    local key='PatchId'

    local patchIds=''
    local patchIds2=''
    if ReadIni "${INI_FILEPATH}" "${section}" "${key}" patchIds; then
        if [ -n "${patchIds}" ]; then
            DeleteFromList "${patchIds}" ',' "${mnfstPatchId}" patchIds2
            if [ "${patchIds}" != "${patchIds2}" ]; then
                WriteIni "${INI_FILEPATH}" "${section}" "${key}" "${patchIds2}"
            fi
        fi
    fi
}

CopySymlink()
{
    cp -TP "$1" "$2" || Die "'cp' failed CopySymlink '$1' '$2' (exit code=$?)"
}

MoveSymlink()
{
    mv -fT "$1" "$2" || Die "'mv' failed MoveSymlink '$1' '$2' (exit code=$?)"
}

RemoveSymlink()
{
    rm -f "$1"
}

ReloadKlnagent()
{
    if [ -x "${SC_KLNAGENT_BIN}" ]; then
        "${SC_KLNAGENT_BIN}" -reload-connectors
    fi
}

RunInstUninstAtDir()
{
    local path="$1"
    local script="$2"

    "${path}/${script}" "${path}/${INSTALL_LOG_FILEPATH}"
    return $?
}

Call_GetUpgradeState()
{
    GetUpgradeState UPGRADE_STATE PATCH_VERSION OLD_VERSION
}

SwitchToUpgradeState()
{
    local newState="$1"
    shift

    Log "Switch upgrade state: '${UPGRADE_STATE}' -> '${newState}': $@"

    "$@"

    Call_GetUpgradeState
    if [ "${UPGRADE_STATE}" != "${newState}" ]; then
        Log "WARNING: GetUpgradeState() returned unexpected upgrade state: ${UPGRADE_STATE}; expected: ${newState}"
    fi
}

ExecLauncher()
{
    local ver="$1"

    Trace "ExecLauncher($1)"

    exec "${CURRENT}/${LAUNCHER_FILEPATH}" || Die "Failed to exec ${ver} launcher (exit code=$?)"
}

DoUninstall()
{
    Trace "DoUninstall()"

    RunInstUninstAtDir "${CURRENT}" "${UNINSTALL_FILEPATH}"
}

DoUpgrade()
{
    Trace "DoUpgrade()"

    if ! IsFanotifyOn; then
        RemoveKernelModuleFiles
    fi

    RunInstUninstAtDir "${UPGRADE}" "${INSTALL_FILEPATH}" || return $?

    if [ -f "${PREVIOUS}${EVENTS_STORAGE_FILEPATH}" ]; then
        cp -fpv "${PREVIOUS}${EVENTS_STORAGE_FILEPATH}" "${UPGRADE}${EVENTS_STORAGE_FILEPATH}" || return $?
    fi

    ReloadKlnagent

    return 0
}

DoStart()
{
    Trace "DoStart()"

    Log "Try to start version '${PRODUCT_VERSION}' at install directory '$(readlink ${CURRENT})'"
}

DoUpgradeFailed()
{
    Trace "DoUpgradeFailed()"

    RunInstUninstAtDir "${FAILED}" "${UNINSTALL_FILEPATH}"
}

DoStartFailed()
{
    Trace "DoStartFailed()"

    RunInstUninstAtDir "${FAILED}" "${UNINSTALL_FILEPATH}"
}

DoRollingBack()
{
    Trace "DoRollingBack()"

    Log "Revoke patch approval at rolling back patch '${PATCH_VERSION}'"
    RevokePatchApproval "${FAILED}"

    if ! IsFanotifyOn; then
        RemoveKernelModuleFiles
    fi

    if ! RunInstUninstAtDir "${CURRENT}" "${INSTALL_FILEPATH}"; then
        EmitOfflineEvent 'PatchRollbackError' 'Error' 'PatchVersion' "${PATCH_VERSION}" 'Reason' ''
    fi

    ReloadKlnagent
}

AssumeInstalled()
{
    Trace "AssumeInstalled()"

    local cmp=''
    if CompareVersions "${OLD_VERSION}" "${PATCH_VERSION}" cmp && [ "${cmp}" = '>' ]; then
        Log "Revoke previously installed patch approval at downgrading from '${OLD_VERSION}' to '${PATCH_VERSION}'"
        RevokePatchApproval "${PREVIOUS}"
    fi

    local prevDirName="$(readlink "${PREVIOUS}")"

    SwitchToUpgradeState 'Installed' RemoveSymlink "${PREVIOUS}"

    Log "Remove previous version of files at '${VAR_DIRECTORY}${prevDirName}'"
    rm -rf "${VAR_DIRECTORY}${prevDirName}"
}

DoUpdateCryptoSsl()
{
    Trace 'DoUpdateCryptoSsl()'

    local current="${CRYPTO_SSL_WORKING_COPY_FILEPATH}"
    local ssl_symlink="${CRYPTO_SSL_SYMLINK_FILEPATH}"

    if [ ! -f "${current}" ]; then
        cp -frp "${CRYPTO_SSL_FILEPATH}" "${current}" || { Trace "Cannot copy ${CRYPTO_SSL_FILEPATH} to ${current}"; return 1; }
        ln -sfT "${current}" "${ssl_symlink}"
        return 0
    fi

    if [ ! -h "${CRYPTO_SSL_SYMLINK_NEW_FILEPATH}" ]; then
        return 0
    fi

    local prev="${CRYPTO_SSL_PREV_FILEPATH}"
    local ssl="$(readlink -e "${ssl_symlink}")"
    local upstream="$(readlink -e "${CRYPTO_SSL_SYMLINK_NEW_FILEPATH}")"

    cmp -s "${current}" "${upstream}"
    local rc=$?

    if [ $rc -eq 0 ]; then
        return 0;
    fi

    if [ ! -f "${prev}" ] || [ "${ssl}" != "$(readlink -e "${prev}")" ]; then
        cp -frp "${ssl}" "${prev}" || { Trace "Cannot copy ${ssl} to ${prev}"; return 1; }
    fi

    cp -frp "${upstream}" "${current}.tmp" || { Trace "Cannot copy ${upstream} to ${current}.tmp"; return 1; }

    mv -f "${current}.tmp" "${current}" || { Trace "Cannot rename ${current}.tmp to ${current}";  return 1; }

    ln -sfT "${current}" "${ssl_symlink}"

    return 0
}

ProcessUpgradeState()
{
    case "${UPGRADE_STATE}" in
        'Installed')
            DoUpdateCryptoSsl
            return 1
            ;;
        'Uninstall')
            if ! IsFanotifyOn; then
                if IsKernelModuleLoaded; then
                    Log "Cannot do 'Uninstall' now: UseFanotify=false and kernel module is loaded: operating system restart is required"
                    return 1
                fi
                if ! PrepareKernelModule "${UPGRADE}" '--build-only'; then
                    Log "Failed to build new kernel module"
                    SwitchToUpgradeState 'RollingBack' MoveSymlink "${UPGRADE}" "${FAILED}"
                    EmitOfflineEvent 'PatchInstallError' 'Error' 'PatchVersion' "${PATCH_VERSION}" 'Reason' 'Failed to build new kernel module'
                    return 0
                fi
            fi

            DoUninstall
            SwitchToUpgradeState 'Upgrade' CopySymlink "${CURRENT}" "${PREVIOUS}"
            ;;
        'Upgrade')
            if DoUpgrade; then
                SwitchToUpgradeState 'Start' MoveSymlink "${UPGRADE}" "${CURRENT}"
                ExecLauncher 'new'
            else
                SwitchToUpgradeState 'UpgradeFailed' MoveSymlink "${UPGRADE}" "${FAILED}"
            fi
            ;;
        'Start')
            DoStart
            return 1
            ;;
        'UpgradeFailed')
            DoUpgradeFailed
            SwitchToUpgradeState 'RollingBack' RemoveSymlink "${PREVIOUS}"
            EmitOfflineEvent 'PatchInstallError' 'Error' 'PatchVersion' "${PATCH_VERSION}" 'Reason' 'Upgrade procedure failed'
            ;;
        'StartFailed')
            DoStartFailed
            SwitchToUpgradeState 'RollingBack' MoveSymlink "${PREVIOUS}" "${CURRENT}"
            EmitOfflineEvent 'PatchInstallError' 'Error' 'PatchVersion' "${PATCH_VERSION}" 'Reason' 'Failed to start application'
            ExecLauncher 'old'
            ;;
        'RollingBack')
            DoRollingBack
            SwitchToUpgradeState 'RolledBack' CopySymlink "${FAILED}" "${UPGRADE}"
            EmitOfflineEvent 'PatchRollbackCompleted' 'Informational' 'PatchVersion' "${PATCH_VERSION}"
            ;;
        *)
            Log "No actions required for upgrade state '${UPGRADE_STATE}'"
            return 1
            ;;
    esac

    return 0
}

Main()
{
    Repeat 85 '='
    Trace "Main(): pid=$$, PPID=$PPID"

    trap "" HUP
    trap "" INT
    trap "Stop" TERM

    echo $$ > "${LAUNCHER_PID}"

    while :; do
        Call_GetUpgradeState || Die "Failed to get upgrade state (exit code=$?): returned state='${UPGRADE_STATE}'"
        Log "Current upgrade state: '${UPGRADE_STATE}'"
        if ! ProcessUpgradeState; then
            break
        fi
    done

    GetBackupDir BACKUP_CURRENT
    local backupState=$?

    local backupPrepared=0
    if [ ${backupState} -ne 0 ]; then
        PrepareBackup
        backupPrepared=$?
    fi

    PopulateDataDirectory

    local wdPid
    local startDetectedFl
    local alreadyRestoredFl=0

    while true; do
        startDetectedFl=0

        Log "Running wdserver..."


        "${WDSERVER_BIN}" ${WDSERVER_FLAGS} -e "${PRODUCT_BIN}"

        while :; do
            sleep 5

            wdPid="$(pidof "${WDSERVER_BIN}")"

            if [ -z "${wdPid}" ]; then
                Log "'wdserver' process is not found"
                Log "Print application process state:\n$(ps -ef | grep -F "${PRODUCT_NAME}")"
                break
            fi

            if IsProductReady; then
                if [ ${startDetectedFl} -eq 0 ]; then
                    startDetectedFl=1

                    if [ "${UPGRADE_STATE}" = 'Start' ]; then
                        AssumeInstalled
                    fi

                    if [ "${backupPrepared}" -eq 0 ]; then
                        backupPrepared=1
                        if CommitBackup; then
                            GetBackupDir BACKUP_CURRENT
                            backupState=$?
                            Log "Backup has been switched to '${BACKUP_CURRENT}' directory (state ${backupState})"
                        fi
                    fi
                fi
            fi
        done

        if [ "${backupPrepared}" -eq 0 ]; then
            if [ ${backupState} -ne 0 ]; then
                Log "Clear prepared but not committed backup"
                backupPrepared=1
                RemoveNewUniqDir
            fi
        fi

        if [ ${startDetectedFl} -eq 0 ] && [ "${UPGRADE_STATE}" = 'Start' ]; then
            Log "******* Failed to start application the first time after upgrade *******"
            SwitchToUpgradeState 'StartFailed' CopySymlink "${CURRENT}" "${FAILED}"
            ProcessUpgradeState
        fi

        if [ ${startDetectedFl} -eq 0 ] && [ ${alreadyRestoredFl} -ne 0 ]; then
            if [ ${backupState} -eq 2 ]; then
                Log "Switch backup to init storage (state 1)"
                BACKUP_CURRENT="${SHARED_INIT_STORAGE_DIRECTORY}"
                backupState=1
                Log "Setting current symlink to '${SHARED_INIT_STORAGE_DIRECTORY}'"
                ln -Tsfv "${SHARED_INIT_STORAGE_DIRECTORY}" "${BACKUP_CURRENT_LNK}"
            else
                if [ ${backupState} -eq 1 ]; then
                    Log "Switch backup to empty (state 0)"
                    BACKUP_CURRENT=''
                    backupState=0
                    Log "unlink current symlink"
                    unlink "${BACKUP_CURRENT_LNK}"
                fi
                Die "Product cannot start with the init config"
            fi
        fi

        RestoreFromBackup "${BACKUP_CURRENT}"
        alreadyRestoredFl=1


    done
}

umask "${DEFAULT_FILE_MODE_CREATION_MASK}" && Main "$@" >>"${LAUNCHER_LOG}" 2>&1
